import sys
import os
from plotconfig import *

def plot_data( outname, datname, fitname, _skiprows, logx, **kwargs ):
  # outname: name for the EPS file that will be generated
  # datname: name of the text file containing exported data 
  # fitname: name of the text file containing fit traces for data (if any)

  #
  # Load cfg file if available.
  #
  cfg = kwargs.get( 'cfgname', None )
  if cfg and cfg.lower() != 'none':
    kin_load_cfg_file( cfg, kwargs, 1 )

  fig, ax = kin_figure_init( figsize="9x7" )
    
  #
  # Load data & plot
  #
  dat = []
  dat_len = 0
  if datname and datname.lower() != "none":
    (dheader,dat) = kin_load_data( datname, _skiprows )
    dat_len = len(dat[0])
    print( "loaded data, dat_len is " + str(dat_len))
    kwargs['zorder'] = 0
    kwargs['plot_sigma'] = kwargs.get( 'plot_sigma', True )
    sigma_avg,sigma_cols = get_sigma_from_header( dheader )
    kwargs['sigmaAvg'] = sigma_avg
    kwargs['sigmaCols'] = sigma_cols

  dminx,dmaxx,dminy,dmaxy = kin_plot( dat, dheader[_skiprows-1], ax, **kwargs )

  #
  # Load fitted traces & plot if so configured.
  #
  if fitname and fitname.lower() != 'none':
    _skiprows = min( _skiprows, 2 )
      # while the data may have had up to three rows of header info if it had
      # sigma values and series values, a fitted trace will never have the sigma.
    (fheader,fdat) = kin_load_data( fitname, _skiprows )
    fit_len = len(fdat[0])
    print( "loaded fits, fit_len is " + str(fit_len))
    kwargs.pop('colors',None)
    kwargs.pop('plot_sigma',None)
    kwargs['color'] = 'k'
    kwargs['marker'] = '-'
    kwargs['linewidth'] = 0.75
    kwargs['zorder'] = 10
    kin_plot( fdat, fheader[_skiprows-1], ax, **kwargs )

  #
  # Format axes & save EPS file. 
  #
  if len(dat) and outname:
    #
    # (optionally?) draw a dashed line at y=0 if dminy<0
    #
    if dminy < 0:
      ax.axhline( y=0, ls='dashed', c='#000000' )

    #
    # set plot boundaries from options if passed
    #
    x0 = kwargs.get( 'x0', dminx )
    x1 = kwargs.get( 'x1', dmaxx )
    y0 = kwargs.get( 'y0', dminy )
    y1 = kwargs.get( 'y1', dmaxy )

    #
    # set axis bounds & format them
    #
    axis1 = kwargs.get( 'x_label', None )
    axis2 = kwargs.get( 'y_label', None )
    expand_x_axis = 'x0' not in kwargs and dat_len < 50 
    expand_y_axis = 'y1' not in kwargs
    kin_format_axes( ax, x0, x1, y0, y1, axis1, axis2, logx, True, expand_x_axis, expand_y_axis )
    
    #
    # save figure as EPS, PNG
    #
    fig.savefig( outname, format='eps', bbox_inches='tight' )
    fig.savefig( outname.replace(".eps",".png"), format='png', bbox_inches='tight', dpi=300 )


#-------------------------------------------------------------------------------------------------


if __name__ == "__main__":

  argc = len( sys.argv )
  if argc <= 2:
    print( "Usage: python plotData.py <eps_filename> <data_filename> <fit_filename> <axis1Label> <axis2Label> <skiprows> <log> <data_marker>" )
    print( "Ex:    python plotData.py burst.eps my.data.txt\n" )
  else:

    print( "\n===\n=== plotData.py\n===\n" );

    skip = 2
    logx = -99
    axis1 = None
    axis2 = None
    data_marker = '.'
    cfgname='none'
    if argc > 1:
      oname = sys.argv[1]
    if argc > 2:
      dname = sys.argv[2]
    if argc > 3:
      fname = sys.argv[3]
    if argc > 4:
      axis1 = sys.argv[4]
    if argc > 5:
      axis2 = sys.argv[5]
    if argc > 6:
      skip = int(sys.argv[6])
    if argc > 7:
      logx = float(sys.argv[7])
    if argc > 8:
      data_marker = sys.argv[8]
    if argc > 9:
      cfgname = sys.argv[9]

    print( "dat is: " + dname )
    print( "fit is: " + fname )
    print( "out is: " + oname )
    print( "axis1 is : " + str( axis1 ) )
    print( "axis2 is : " + str( axis2 ) )
    print( "skip is: " + str(skip) )
    print( "logx is: " + str(logx) )
    print( "dmark is: " + data_marker )

    if not dname or not os.path.isfile( dname ):
      print( "The data file '" + dname + "' is not found.\n" )
      quit()

    # Really we only want to pass the custom data marker if it is not the default of '.'
    # This way the plot routine will choose the marker and size based on amount of data.
    kwargs = {}
    if data_marker != ".":
      kwargs['marker'] = data_marker

    kwargs['x_label'] = axis1
    kwargs['y_label'] = axis2
      # allow these to still get passed on cmdline, put put them into kwargs.

    kwargs['cfgname'] = cfgname

    plot_data( oname, dname, fname, skip, logx, **kwargs )

    print( "done.\n" )

