import sys
import os
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np
import math
from plotconfig import *

def _plots_layout_with_optional_residuals( figsize, **kwargs ):
  # layout a figure with two subplots, a main data+sim plot, and a smaller plot 
  # up top that shares the x-axis and displays the residuals.

  if not kwargs.get( 'plot_residuals', None ):
    fig, exp = kin_figure_init( figsize, True, **kwargs )
    return fig, exp, None

  #
  # To layout the plots, I'm going to make an assumption that for a 9x7 exp plot,
  # we'll want a 2x7 residuals display, such that the total figure is square.
  #
  print( "create fig with figsize = " + str(figsize) )
  fig = kin_figure_init( figsize, False, **kwargs )

  padx = .1
  pady = .02

  plot_rect_exp = [ 0+padx, 0+pady  , 1-padx, 7.0/9.0-pady ]
  plot_rect_res = [ 0+padx, 7.0/9.0+pady, 1-padx, 2.0/9.0-pady ]

  print( "create exp plot with plot_rect_exp = " + str(plot_rect_exp) )
  exp = fig.add_axes( plot_rect_exp )
  res = fig.add_axes( plot_rect_res, sharex=exp )
  plt.setp( res.get_xticklabels(), visible=False )

  return fig, exp, res


def plot_experiment( outname, simname=None, datname=None, fitname=None, fsbname=None, svdname=None, axis1=None, axis2=None, _skiprows=1, logx=-99, **kwargs ):
  # outname: name for the EPS file that will be generated
  # simname: name of the text file containing an exported simulation
  # fitname: name of the text file containing fit traces (if any)
  # datname: name of the text file containing exported data (if any)
  # sim or dat may be omitted to plot data-only or simulation-only

  kwargs['line_brightness'] = kwargs.get( 'line_brightness', 0.7 )
  kwargs['data_lines'] = kwargs.get( 'data_lines', 0 )
  kwargs['linewidth'] = kwargs.get( 'linewidth', 0.75 )
    # these are both options that when exporting a simulation, are covered by
    # the export options panel, but we need to ensure they are set to something
    # in case this is being called from an analytic fit.

  #
  # Load cfg file if available.
  #
  cfg = kwargs.get( 'cfgname', None )
  if cfg and cfg.lower() != 'none':
    kin_load_cfg_file( cfg, kwargs, 1 )
    axis1 = kwargs.get( 'x_label', axis1 )
    axis2 = kwargs.get( 'y_label', axis2 )


  #
  # Load data
  #
  dat = []
  dat_len = 0
  if datname and datname.lower() != "none":
    (dheader,dat) = kin_load_data( datname, _skiprows )
    dat_len = len(dat[0])
    print( "loaded data, dat_len is " + str(dat_len))


  #
  # Load sim
  #
  sim = []
  sim_cols = 0
  if simname and simname.lower() != "none":
    (sheader,sim) = kin_load_data( simname, _skiprows )
    sim_cols = sim.shape[0]

  #
  # Load fit
  #
  fit = []
  fit_cols = 0
  if fitname and fitname.lower() != "none":
    (fheader,fit) = kin_load_data( fitname, _skiprows )
    fit_cols = fit.shape[0]

  #
  # Load observable traces at fitspace bounds (fsbounds) if available
  #
  fsb = []
  fsb_cols = 0
  if fsbname and fsbname.lower() != "none":
    (fsbheader,fsb) = kin_load_data( fsbname, _skiprows )
    fsb_cols = fsb.shape[0]
    # print( "checking time cols for " + fsbname + " ..." )
    # kin_check_times_ascending( fsb, fsbheader[_skiprows-1] )


  #
  # Load svd inverse-transform amplitudes
  #
  svd = []
  svd_cols = 0
  if svdname and svdname.lower() != "none":
    (svdheader,svd) = kin_load_data( svdname, _skiprows )
    svd_cols = svd.shape[0]
    print( "svd_cols is %d" % svd_cols )


  #
  # Load residuals
  #
  resu = res = None
  resname = simname.replace( ".txt", ".resu.txt" )
  if "resu" in resname and os.path.isfile( resname ):
    (rheader,resu) = kin_load_data( resname, _skiprows )
      # Load the un-normalized residuals

    resname = simname.replace( ".txt", ".res.txt" )
    if os.path.isfile( resname ):
      (rheader2,res) = kin_load_data( resname, _skiprows )


  #
  # Setup the figure, with optional residuals subplot if we have residuals.
  #
  kwargs['plot_residuals'] = kwargs.get( 'plot_residuals', None ) and resu is not None
  figsize = '9x9' if kwargs['plot_residuals'] else '9x7'
  fig, ax, ax_res = _plots_layout_with_optional_residuals( figsize, **kwargs )

  #
  # Plot data under sim, unless dat_len < 50
  # 
  dminx = dminy = sminx = sminy = sys.float_info.max
  dmaxx = dmaxy = smaxx = smaxy = sys.float_info.min

  saved_linewidth = kwargs['linewidth']
  saved_line_brightness = kwargs['line_brightness']
    # because we may set it to a prescribed value if we are plotting data as lines.

  if fsb_cols > 0:
    # This needs to be printed first underneath all the other data.  Because it's 
    # getting unwieldly to set and unset arguments for each pass of plotting, I'm
    # just going to copy and modify the args.
    fsb_args = dict( kwargs )
    #fsb_args['marker'] = '-'
    # kin_plot( fsb, fsbheader[0], ax, **fsb_args )
    #fsb_args['linewidth'] = 0.25
      # a thin line looks good when it is a darker color, but I want to try using the same
      # line width as the traces will use, which is what the program does, because otherwise
      # the fsb traces may not protrude out from under the fitted trace.
    fsb_args['facecolor'] = '0.75'
    fsb_args['edgecolor'] = '0.75'
    kin_plot_fill_same_labels( fsb, fsbheader[_skiprows-1], ax, **fsb_args )


  #
  # local copy of kwargs modified for data/residual plotting based on options
  #
  dat_args = dict( kwargs )
  if dat_args['data_lines'] == 1:
    dat_args['marker'] = '-'
    dat_args['linewidth'] = 0.5
    dat_args['line_brightness'] = 1
  kwargs['marker'] = '-'
    # always plot fit/sim traces as lines.
  if dat_len != 0:
    sigma_avg,sigma_cols = get_sigma_from_header( dheader )
    dat_args['sigmaAvg'] = sigma_avg
    dat_args['sigmaCols'] = sigma_cols

  kwargs['color_indices'] = kwargs.get( 'sim_indices', None );
  dat_args['color_indices'] = kwargs.get( 'data_indices', None )

  #
  # if lots of data, plot data underneath fit/sim traces, else vice-versa.
  #
  if dat_len > 50:
    dminx,dmaxx,dminy,dmaxy = kin_plot( dat, dheader[_skiprows-1], ax, **dat_args )
    if sim_cols > 0:
      sminx,smaxx,sminy,smaxy = kin_plot( sim, sheader[_skiprows-1], ax, **kwargs )
  else:
    if sim_cols > 0:
      if dat_len == 0 and fit_cols > 0:
        # if we are plotting a simulation with fitted traces on top, make the 
        # sim traces fatter.
        kwargs['linewidth'] = kwargs.get( 'linewidth', 0.75 ) * 2.0
      sminx,smaxx,sminy,smaxy = kin_plot( sim, sheader[_skiprows-1], ax, **kwargs )
    if dat_len > 0:
      dminx,dmaxx,dminy,dmaxy = kin_plot( dat, dheader[_skiprows-1], ax, **dat_args )


  #
  # plot any fit traces last
  #
  if fit_cols > 0:
    kwargs['line_brightness'] = 0.1
      # dark lines for fitted traces, otherwise they'll be colored as normal for observable indices.
    if dat_len == 0:
      # we are plotting fitted lines over simulated lines, so ensure our linewidth is smaller
      # so as not to completely obscure the simulated traces.
      kwargs['linewidth'] = kwargs.get( 'linewidth', 0.75 ) / 2.0
    sminx,smaxx,sminy,smaxy = kin_plot( fit, fheader[_skiprows-1], ax, **kwargs )
      # note we capture the min/max again.  If there was sim data, these values
      # should be the same, but in the case of analytic fits from exp editor,
      # there may not be any sim data, so we need these min/max here.

  if ax_res:
    # plot residuals.  If we have normalized residuals, prefer those.
    # (I think if you have one you'll always have the other)
    residuals = res if res is not None and kwargs['norm_residuals'] else resu
    dat_args['color_indices'] = kwargs.get( 'res_indices', None )
    rminx,rmaxx,rminy,rmaxy = kin_plot( residuals, rheader[_skiprows-1], ax_res, **dat_args )
    kin_thin_ticklabels( ax_res, 6, thinx=False, thiny=True, keep_blank_ticks=True )
    res_label = 'r/$\sigma$' if kwargs['norm_residuals'] else 'r'
    ax_res.set_ylabel( res_label, labelpad=2 )


  if svd_cols > 0:
    kwargs['marker'] = '.'
    kwargs['color'] = '0.5'
    kin_plot( svd, svdheader[_skiprows-1], ax, **kwargs )


  #
  # Format axies & save EPS file. 
  #
  if (len(dat) or len(sim)) and outname:
    #
    # set axis bounds & format them 
    #
    minx = min( dminx, sminx )
    maxx = max( dmaxx, smaxx )
    miny = min( dminy, sminy )
    maxy = max( dmaxy, smaxy )

    x0 = kwargs.get( 'x0', minx )
    x1 = kwargs.get( 'x1', maxx )
    y0 = kwargs.get( 'y0', miny )
    y1 = kwargs.get( 'y1', maxy )
      # axis boundaries passed in the config file can override data min/max values.

    expandY = 'y1' not in kwargs;
      # if explicit y axis boundaries were passed, don't expand these boundaries

    #
    # (optionally?) draw a dashed line at y=0 if y0<0
    #
    if y0 < 0:
      ax.axhline( y=0, ls='dashed', c='#000000' )


    kin_format_axes( ax, x0, x1, y0, y1, axis1, axis2, logx, True, False, expandY )
      # the last two par

    #
    # save figure as EPS, PNG
    #
    fig.savefig( outname, format='eps', bbox_inches='tight' )
    fig.savefig( outname.replace(".eps", ".png"), format='png', bbox_inches='tight', dpi=300 )


  #
  # Plot a histogram of residuals if residuals are available.  
  #
  residuals = resu if resu is not None else res
  
  if residuals is not None:
    fig, ax = kin_figure_init( '9x9', True, **kwargs )
    all_residuals = kin_concat_data( residuals, rheader[_skiprows-1] )

    all_residuals = all_residuals[~np.isnan(all_residuals)]
      # remove NaN entries

    sse = np.sum( all_residuals**2 )
    observed_sigma_wrt_fit = math.sqrt( sse / (len(all_residuals)-1) )
    observed_mean_wrt_fit = np.sum( all_residuals ) / len(all_residuals)

    # experimental - remove residuals that are outside of 6sigma
    s_min = -6 * observed_sigma_wrt_fit
    s_max = 6 * observed_sigma_wrt_fit
    r_min = np.amin( all_residuals )
    r_max = np.amax( all_residuals )
    truncated_residuals = False
    if r_min < s_min or r_max > s_max:
      # if residuals exist outside of 6sigma, remove them.
      all_residuals = all_residuals[(all_residuals >= s_min) & (all_residuals <= s_max)]
      r_min = np.amin( all_residuals )
      r_max = np.amax( all_residuals )
      truncated_residuals = True


    nbins = len(all_residuals) / 10
    nbins = max( 10, nbins )
    nbins = min( 40, nbins )
    w = np.zeros_like( all_residuals ) + 1.0 / len(all_residuals)

    (n,bins,patches) = ax.hist( all_residuals, bins=nbins, weights=w )
      # the weights cause the total sum "height" of the histogram to equal 1,
      # such that the height of any bar gives the fraction of observations
      # in that bar.

    #
    # We're going to center the plot on x=0 regardless of the location of the
    # histogram, so find the domain of the plot.  We'll use a linspace to plot
    # the normpdf against. We'll also scale the normpdf such that the integral
    # of the normpdf equals the integral of the histogram, so they'll appear
    # in the same y-scale.
    #

    r_absmax = max( abs(r_min), abs(r_max) ) * 1.1
    plot_domain = np.linspace( -r_absmax, r_absmax, 100 )
    normpdf_scale = bins[1]-bins[0]

    #
    # If we have a sigma that was associated with the data when it was fit,
    # plot what a normal pdf using this sigma.
    #
    # TODO: I removed this computation above june 28 2018.  See the commit log.
    # I think we should pass in the sigma if we want to include it in the histogram.
    #
    # if sigma != 1.0:
    #   normpdf = mlab.normpdf( plot_domain, 0, sigma )
    #   sigma_str = '%.2E' % sigma
    #   ax.plot( plot_domain, normpdf_scale * normpdf, 'r-', linewidth=1, label='$\sigma_{data}$ = ' + sigma_str )
    dataSigmas = kwargs.get( 'data_sigmas', None )
    if dataSigmas:
      for sigma in dataSigmas:
        if sigma != 1:
          normpdf = mlab.normpdf( plot_domain, 0, sigma )
          sigma_str = '%.2E' % sigma
          ax.plot( plot_domain, normpdf_scale * normpdf, 'r-', linewidth=1, label='$\sigma_{data}$ = ' + sigma_str )




    #
    #  Plot the observed sigma w.r.t. the fit
    #
    normpdf0 = mlab.normpdf( plot_domain, 0, observed_sigma_wrt_fit )
    sigma_str = '%.2E' % observed_sigma_wrt_fit
    ax.plot( plot_domain, normpdf_scale * normpdf0, 'c-', linewidth=1, label='$\sigma_{fit} = $' + sigma_str )

    #
    # we plotted curves to 10% on either side of the histogram.  Set the
    # x-axis limits to these values.
    #
    ax.set_xlim( [-r_absmax,+r_absmax] )

    #
    # legend for normal pdf plots
    #
    legend = ax.legend( loc='best', shadow=None, fontsize=8 )
    legend.get_frame().set_linewidth( 0.5 )

    #
    # Axes labels.  If we truncated the residuals to 6sigma, indicate this.
    #
    xlabel = "Residual"
    if truncated_residuals:
      xlabel = "Residual [$-6\sigma,+6\sigma$]"
    ax.set_xlabel( xlabel )
    ax.set_ylabel( "Frequency" )

    #
    # Save an EPS file, and a 300-dpi PNG file.
    #
    reshistname = simname.replace( ".txt", ".res.hist.eps" )
    reshistname = reshistname.replace( "Text & cfg files", "Histograms" )
      # the last line is a hack after Ken decided he wanted three different locations for
      # different files - the eps/png of the standard export, the txt and cfg files of data, 
      # and the eps/png files of histograms.  Ugh.
    fig.savefig( reshistname.replace(".eps",".png"), format='png', bbox_inches='tight', dpi=300 )
    fig.savefig( reshistname, format='eps', bbox_inches='tight' )
      # note: moving the eps save last because this is now giving a strange error
      # on matplotlib in python 3.5 anaconda - works fine in python 2.7 osx system install






#===============================================================================================

from matplotlib import __version__
#debug

if __name__ == "__main__":
  print( "python version " + str( sys.version_info ))
  print( "matplotlib " + str( __version__ ) )

  argc = len( sys.argv )
  if argc <= 3:
    print( "Usage: python plotExperiment.py <eps_filename> <sim_filename> <data_filename> <fit_name> <axis1Label> <axis2Label> <series_flag (0 or 1)> <log_min_exponent>" )
    print( "Ex:    python plotExperiment.py burst.eps burst.sim.txt burst.data.txt\n" )
    print( "       If you want to omit the sim or the data, use the word none in place of a filename.\n" )
  else:
    print( "\n===\n=== plotExperiment.py\n===\n" );

    skip = 1
    logx = -99
    axis1 = None
    axis2 = None
    fsbname = None
    cfgname = None
    svdname = None
    if argc > 1:
      oname = sys.argv[1]
    if argc > 2:
      sname = sys.argv[2]
    if argc > 3:
      dname = sys.argv[3]
    if argc > 4:
      fname = sys.argv[4]
    if argc > 5:
      axis1 = sys.argv[5]
    if argc > 6:
      axis2 = sys.argv[6]
    if argc > 7:
      skip = int(sys.argv[7])
    if argc > 8:
      logx = float(sys.argv[8])
    if argc > 9:
      fsbname = sys.argv[9]
    if argc > 10:
      svdname = sys.argv[10]
    if argc > 11:
      cfgname = sys.argv[11]

    print( "sim is: " + sname )
    print( "dat is: " + dname )
    print( "fit is: " + fname )
    print( "fsb is: " + fsbname )
    print( "svd is: " + svdname )
    print( "out is: " + oname )
    print( "axis1 is : " + str( axis1 ) )
    print( "axis2 is : " + str( axis2 ) )
    print( "skip is: " + str(skip) )
    print( "logx is: " + str(logx) )

    if (not sname or sname.lower() == "none") and (not dname or dname.lower() == "none"):
      print( "You must supply at least one sim or data input file.\n" )
      quit()

    if sname and sname.lower() != "none" and not os.path.isfile( sname ):
      print( "The simulation file '" + sname + "' is not found.  Plot will contain no simulation.\n" )
    
    if dname and dname.lower() != "none" and not os.path.isfile( dname ):
      print( "The data file '" + dname + "' is not found.  Plot will not contain data.\n" )

    #
    # Note that we use the series flag to tell us how many header rows to
    # skip when reading the output files.  Normally, one header line exists,
    # but in the case of a series experiment, there are two header lines.
    #
    plot_experiment( oname, sname, dname, fname, fsbname, svdname, axis1, axis2, skip, logx, cfgname=cfgname )

    print( "done.\n" )

