import sys
import os
import numpy as np
import math
import re
from plotconfig import *
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.colors as mplcolors
import matplotlib.colorbar as mplcolorbar


#--------------------------------------------------------------------------------------------------

def _plots_layout_rectangular( numcols, numplots, plotsize, w_pad, sharey=True ):
  # if sharey, plots in the same row will share a single y axis
  # This can be used for the "1D" chi2 plots in fitspace, as well
  # as the normalized forward rate plots.

  pw,ph = wh_from_figsize( plotsize )

  numrows = int( math.ceil( float(numplots) / numcols ) )
  figw = numcols * pw
  figh = numrows * ph
  figsize = figsize_from_wh( figw, figh )

  fig = kin_figure_init( figsize=figsize, init_single_plot=False )  

  plotcount = 0
  plots = []
  for r in range( numrows ):
    plots.append( [] )
    row = plots[-1]
    for c in range( numcols ):

      shared_axis = None
      if sharey and c>0:
        shared_axis = row[0]

      plotcount += 1
      row.append( fig.add_subplot( numrows, numcols, plotcount, sharey=shared_axis ) )

      if plotcount == numplots:
        break

  fig.tight_layout( pad=2.0, h_pad=2.0, w_pad=w_pad )
  return fig,plots

#--------------------------------------------------------------------------------------------------

def _plots_layout_triangular( numparams, plotsize, **kwargs ):
  # This is specifically for 2D/3D fitspace plots, in which we display a 
  # lower-triangular matrix of plots, in which rows and columns both share
  # axes.
  #
  # We create the subplots via calls to axes() because this allows subplots to
  # be placed arbitrarily on the figure by passing a rect.  Otherwise add_subplot
  # would require us to create the full square matrix of plots, and just ignore
  # those in the upper triangle.
  #
  # To be consistent with row,col addressing of subplots however, we'll still return
  # an array of arrays indexable as plot[row][col].
  #
  # Actually, right at the end I'm going to linearize the array of plots, so that
  # a calling function that actually does the plotting doesn't need to know about
  # the layout of the plots.

  pw,ph = wh_from_figsize( plotsize )
  numcols = numrows = int( numparams ) - 1

  pad_cm = 2.0
  pad_interior_cm = .5

  figw = math.ceil( numcols * pw + (numcols-1)*pad_interior_cm + pad_cm * 2 )
  figh = math.ceil( numrows * ph + (numrows-1)*pad_interior_cm + pad_cm * 2 )
  
  figw = figw + pw
    # add width for an additional column to make room for a colorbar
  
  figsize = figsize_from_wh( int(figw), int(figh) )

  fig = kin_figure_init( figsize=figsize, init_single_plot=False, font_size=16, ticklabel_size=12 ) 
  proj = '3d' if kwargs['plot_type'] == 'surface' else None

  padding = pad_cm / figw
  padding_interior = pad_interior_cm / figw
    # padding expressed as fraction of figure size
  plotsize_x = pw / figw
  plotsize_y = ph / figh
  plotinc_x = plotsize_x + padding_interior
  plotinc_y = plotsize_y + padding_interior
  start_x = 0+padding
  start_y = 1.0-padding-plotsize_y
  plot_rect = [ start_x, start_y, plotsize_x, plotsize_y ]
    # start with the upper-left plot

  # Because we want all plots in a row to share a y-axies, and also all
  # plots in a column to share an x-axis, it's easiest to create the
  # edge plots first, so that we can refer to them when creating the
  # interior plots that want to share the axes.  Actually, we need
  # to create the bottom row first, since all columns share those a-axis.

  plots = []
  currentcols = 0
  for r in range( numrows ):
    plots.append( [] )
      # append an array for each row, and increment the plot_rect to the next row
    if r == numrows-1:
      # on the last row, create the bottom-edge plots.  These all share
      # the same y-axis, and will each be the 'master' for shared x-axis.
      lastrow = plots[-1]
      for c in range( numcols ):
        shared_y = lastrow[0]['ax'] if c>0 else None
        #print( "creating bottom edge plot for r,c " + str([r,c]) )
        lastrow.append( { 'sharex':None, 'sharey': shared_y, 'ax':fig.add_axes(plot_rect,sharey=shared_y,projection=proj) } )
        plot_rect[ 0 ] += plotinc_x
    plot_rect[ 1 ] -= plotinc_y


  # Now create the left-edge + interior plots.
  #
  plot_rect[0] = start_x
  plot_rect[1] = start_y
  for r in range( numrows-1 ):
    row = plots[r]
    for c in range( r+1 ):
      shared_x = lastrow[c]['ax']
      shared_y = row[0]['ax'] if c>0 else None
      #print( "creating left-edge/interior plot for r,c " + str([r,c]) + " shared_x: " + str(shared_x) )
      row.append( { 'sharex':shared_x, 'sharey': shared_y, 'ax':fig.add_axes(plot_rect, sharex=shared_x, sharey=shared_y, projection=proj) } )
      plot_rect[0] += plotinc_x
    plot_rect[0] = start_x
    plot_rect[1] -= plotinc_y

  # Finally, create a colorbar that will be shared by all of these plots.
  #
  plot_rect[0] = start_x + plotinc_x * numcols + plotinc_x/4
  plot_rect[1] = start_y - plotinc_y * (numrows-1)
  plot_rect[2] = plotinc_x / 5
  plot_rect[3] = plotinc_y * numrows
  colorbar_axes = fig.add_axes( plot_rect )


  # New - return a linear array of plots instead, in the order that they should be plotted
  # based on the order that the grids are output from the c++ software.
  #
  plots_linear = []
  for c in range(numcols):
    for r in plots:
      if len(r) > c:
        plots_linear.append( r[c] )





  return fig,plots_linear,colorbar_axes

#--------------------------------------------------------------------------------------------------

def _plot_fitspace_1d( outname, datname, **kwargs ):
  # The .fsp file contains lots of information, and we need to scan
  # to the section that is titled 1D TABLES and then do a plot for
  # each of the tables we find there.

  # minx = miny = sys.float_info.max
  # maxx = maxy = sys.float_info.min

  param_count = int(kwargs['PARAMCOUNT'])
  chi2thresh  = float(kwargs.get('CHI2THRESH',1))
  chi2bound   = float(kwargs.get('CHI2BOUND',1))
  desired_cols = 3 if param_count > 4 else 2
  fig,plots = _plots_layout_rectangular( desired_cols, param_count, '5x5', -2.5, True ) 

  processing_tables = False;
  plot_index = 0
  global_miny = 1
  global_maxy = 0
  with open(datname) as f:
    for index, line in enumerate(f):
      if '1D TABLES' in line:
        processing_tables = True

      if '2D TABLES' in line:
        break

      if processing_tables and 'Chi2' in line:
        header,dat = kin_load_data( datname, index+1, True, True )
        h = header[-1][0].split()
        row = plot_index // desired_cols
        col = plot_index % desired_cols
        subplot = plots[row][col]

        ylabel = "$minChi^2/Chi^2$" if col==0 else None
        xlabel = parameter_label( h[0] )

        dat[0],xlabel,factor = kin_normalize_axis( dat[0], xlabel )
          # factor out into label if minx,maxx are very large

        minx,maxx,miny,maxy = kin_plot( dat, [], subplot, color='b' ) 
        global_miny = min( global_miny, miny )
        global_maxy = max( global_maxy, maxy )

        subplot.axhline( y=chi2bound, ls='dashed', c='#606060' )

        subplot.axhline( y=1, ls='solid', c='#000000', linewidth=0.5 )

        if chi2bound < miny:
          miny = chi2bound

        kin_format_axes( subplot, minx, maxx, miny, maxy, xlabel, ylabel, -99, False, True )  

        kin_thin_ticklabels( subplot, 8, True, False  )
          # if there are too many ticklabels they overlap, so thin them

        if col != 0:
          plt.setp(subplot.get_yticklabels(), visible=False)
        plot_index += 1


  # set all plots to same y-scale
  # HACKY.  To do this better we'd need to refactor the above to read in all
  # 1d data first so that we can find global min/max before creating each plot.
  for i in range( plot_index ):
    row = i // desired_cols
    col = i % desired_cols
    subplot = plots[row][col]
    deltay = (global_maxy - global_miny) * 0.1
    subplot.set_ylim( [global_miny-deltay, global_maxy+deltay] )

          
  fig.savefig( outname, format='eps' )
  fig.savefig( outname.replace(".eps",".png"), format='png', dpi=300)

#--------------------------------------------------------------------------------------------------

def _plot_fitspace_normrates( outname, datname, **kwargs ):

  param_count = 0
  with open(datname) as f:
    for index, line in enumerate(f):
      if 'Normalized fitted' in line:
        param_count += 1


  desired_cols = 1
  fig,plots = _plots_layout_rectangular( desired_cols, param_count, '9x9', .001, True ) 

  processing_tables = False;
  plot_index = 0
  with open(datname) as f:
    for index, line in enumerate(f):
      if 'NORMALIZED' in line:
        processing_tables = True

      if processing_tables and 'Normalized fitted' in line:
        header,dat = kin_load_data( datname, index+3, True, True )
        row = plot_index // desired_cols
        col = plot_index % desired_cols
        subplot = plots[row][col]

        ylabel = None
        xlabel = parameter_label( header[-1][0] )

        dat[0],xlabel,factor = kin_normalize_axis( dat[0], xlabel )
          # factor out into label if minx,maxx are very large

        minx,maxx,miny,maxy = kin_plot( dat, header[-1], subplot )  
        legend = subplot.legend( loc='best', shadow=None, fontsize=8 )
        legend.get_frame().set_linewidth( 0.5 )

        miny = 0
        maxy = 1

        kin_format_axes( subplot, minx, maxx, miny, maxy, xlabel, ylabel, -99, False, True )  

        kin_thin_ticklabels( subplot, 8, True, False  )
          # if there are too many ticklabels they overlap, so thin them

        if col != 0:
          plt.setp(subplot.get_yticklabels(), visible=False)
        plot_index += 1

  if False and kwargs.get( "test" ):
    fig.show()
  else:
    fig.savefig( outname, format='eps' )
    fig.savefig( outname.replace(".eps",".png"), format='png', dpi=300)


#--------------------------------------------------------------------------------------------------

def _plot_ptype( ptype, ax, p1, p2, dat, xlabel, ylabel, colormap, **kwargs ):

  vmin=kwargs.get('vmin',0)
  vmax=kwargs.get('vmax',1)

  c_levels_n = kwargs.get( 'contour_n', 32 )  
  contour_levels = kwargs.get( 'contour_levels', None )

  c = None
  if ptype=="contour":
    c = ax.contour( p1, p2, dat, c_levels_n, levels=contour_levels, cmap=colormap )

  elif ptype=="contourf":
    c = ax.contour( p1, p2, dat, c_levels_n, levels=contour_levels, linewidths=0.1, colors='k' )
    c = ax.contourf( p1, p2, dat, c_levels_n, levels=contour_levels, vmin=vmin, vmax=vmax, cmap=colormap )

  elif ptype=="contourf1024":
    # note, not contour with lines
    c = ax.contourf( p1, p2, dat, c_levels_n, levels=contour_levels, vmin=vmin, vmax=vmax, cmap=colormap )

  elif ptype=="surface":
    X,Y = np.meshgrid( p1, p2 )
    c = ax.plot_surface( X, Y, dat, rstride=1, cstride=1, vmin=vmin, vmax=vmax, cmap=colormap )

  elif ptype=="surface2d":
    X,Y = np.meshgrid( p1, p2 )
    c = ax.pcolor( X, Y, dat, vmin=vmin, vmax=vmax, cmap=colormap )

  minx = np.amin( p1 )
  maxx = np.amax( p1 )
  miny = np.amin( p2 )
  maxy = np.amax( p2 )
    # scan_grids already normalized axes & labels & adjusted the p1/p2 ranges
  kin_format_axes( ax, minx, maxx, miny, maxy, xlabel, ylabel, -99, False, False, False ) 

  kin_thin_ticklabels( ax, 8, True, True, False  )
    # if there are too many ticklabels they overlap, so thin them

  plt.setp( ax.get_xticklabels(), visible=xlabel!=None )
  plt.setp( ax.get_yticklabels(), visible=ylabel!=None )

  if ptype=="surface":
    ax.view_init( 18, -130 )

  return c

#--------------------------------------------------------------------------------------------------

def _config_colorbar( axes, **kwargs ):
  contour_levels = None
  contour_n = kwargs.get( 'contour_n', 32 )
  bounds = None
  if 'CHI2YELLOW' in kwargs:
    boundaryValue = float(kwargs['CHI2YELLOW'])
    boundaryColor = kwargs['colormap']( boundaryValue )
    print( " chi2yellow val, color -> " + str(boundaryValue) + ", " +str(boundaryColor) )
  
    # contours_high = np.linspace( boundaryValue, kwargs['vmax'], 8 )
    # contours_low = np.linspace( kwargs['vmin'], boundaryValue, 16 )
    # contour_boundary = [ boundaryValue - ( contours_high[1]-boundaryValue) ]
      # create a special boundary that contains the boundary valued points, in 
      # an attempt to create yellow band here.
    #contour_levels = np.unique( np.concatenate( (contours_low, contour_boundary, contours_high) ) )
      # The difficulty with this technique is that it creates contours that are different
      # sizes for the above-boundary and below-boundary contours, violating a principle of
      # contour plots that all contours be the same size s.t. line-spacing is informative.

    # We want the boundaryValue to be one of the contours (a lower limit).  And we want the 
    # vmax to be the upper bound on the last contour.  Or does that matter?  We could say,
    # we want 24 contours, and one of them needs to be boundaryValue.  Then manually set 
    # the upper end to be +inf, and the lower end to be 0.
    contour_levels = np.linspace( kwargs['vmin'], kwargs['vmax'], contour_n )
    boundIdx = ( np.abs(contour_levels - boundaryValue) ).argmin()
    if boundIdx == len(contour_levels) - 1:
      boundIdx -= 1
    contour_levels -= contour_levels[boundIdx] - boundaryValue
    contour_levels[0] = kwargs['vmin']
    contour_levels[-1] = kwargs['vmax']
    contour_n = len( contour_levels )
#   print( " using contour_levels  -> " + str(contour_levels) )
#   print( " first,main,last sizes -> " + str([contour_levels[1]-contour_levels[0],contour_levels[2]-contour_levels[1],contour_levels[-1]-contour_levels[-2]]) )
    if kwargs.get( 'contour_bounds', True ):
      bounds = contour_levels
#     print( "setting colorbar boundaries to contour_levels...")
        # passing bounds to the ColorbarBase constructor will display ONLY the contour colors.
    ticks = [.25, .5, 1/1.5, 1/1.3, boundaryValue, 1.0 ]

    #labels = [ '4x', '2x', '1.5x', '1.3x', "%.2fx" % (1.0/boundaryValue), '' ]
    labels = [ "0.25", "0.50", "0.67", "%.2f" % (1.0/1.3),"%.2f" % boundaryValue, '']

  cb_cfg = {}
  cb_cfg['contour_levels'] = contour_levels
  cb_cfg['contour_n'] = contour_n
  cb_cfg['colorbar_bounds'] = bounds

  norm = mplcolors.Normalize( vmin=kwargs['vmin'], vmax=kwargs['vmax'] )

  cb = mplcolorbar.ColorbarBase( axes, cmap=kwargs['colormap'], norm=norm, ticks=ticks, boundaries=bounds )
  cb.set_ticklabels( labels )
  return cb,cb_cfg

def _plot_fitspace_2d( outname, grids, **kwargs ):
  param_count = int(kwargs['PARAMCOUNT'])
  ptype = kwargs['plot_type']
  layout = kwargs['layout']

  # TODO: use the layout param to get different layouts
  fig,plots,cbax = _plots_layout_triangular( param_count, '5x5', **kwargs ) 

  # colobar axes, colormap and range 
  kwargs['colorbar_axes'] = cbax
  kwargs['colormap'] = kwargs.get( 'colormap', plt.get_cmap( 'jet' ) )
  kwargs['vmin'],kwargs['vmax'] = get_grids_minmax( grids )
  kwargs['vmin'] = 0
    # we want contours and colors to stretch down to 0, even if the data does not.
  cb,cb_cfg = _config_colorbar( cbax, **kwargs )
  kwargs.update( cb_cfg )


  for i,p in enumerate(plots):
    p1,p2,dat,xlabel,ylabel = get_grid_info( grids, i )

    ax = p['ax']
    xlabel = xlabel if p['sharex'] == None else None
    ylabel = ylabel if p['sharey'] == None else None
      # we'll turn off labels/ticks if we are sharing an axis from another plot   
    c = _plot_ptype( ptype, ax, p1, p2, dat, xlabel, ylabel, **kwargs )

  if kwargs.get( "test" ):
    fig.show()
  else:
    fig.savefig( outname, format='eps' )
    fig.savefig( outname.replace(".eps",".png"), format='png', dpi=300)


#--------------------------------------------------------------------------------------------------

def get_fs_colormap( filename ):
  h,palette = kin_load_data( filename, 3, import_cols=False, break_on_blank_line=False, force_float=True, delimiter=',' )
  #print( "palette is:\n " + str(palette) )
  return mplcolors.ListedColormap( palette )

#--------------------------------------------------------------------------------------------------

def scan_grids( datname, normalize=True):
  params = re.compile( "^(\S+) vs (\S+)")
  p1 = p2 = dat = param1 = param2 = None
  grids = []
  with open(datname) as f:
    for index, line in enumerate(f):
      m = params.match( line )
      if m:
        param1 = m.group(1)
        param2 = m.group(2)
        p1label = parameter_label( param1 )
        p2label = parameter_label( param2 )
        p2,p1,dat = kin_load_data_3d( datname, False, index+3, True, missing_values='NC', filling_values='0' )

        factor = 1
        if normalize:
          p1,p1label,p1factor = kin_normalize_axis( p1, p1label )
          p2,p2label,p2factor = kin_normalize_axis( p2, p2label )

        print( "Found grid %s vs %s" % (param1, param2) )

        grids.append( { 'param1':param1, 'param2':param2, 'xlabel':p2label, 'ylabel':p1label, 'xdat':p2, 'ydat':p1, 'zdat':dat } )

  return grids

def get_grid_info( grids, index ):
  g = grids[index]
  return ( g['xdat'], g['ydat'], g['zdat'], g['xlabel'], g['ylabel'] )

def get_grids_minmax( grids ):
  mindat = sys.float_info.max
  maxdat = sys.float_info.min
  for i,g in enumerate(grids):
    p1,p2,dat,xlabel,ylabel = get_grid_info( grids, i )
    dmin = np.amin( dat )
    dmax = np.amax( dat )
    mindat = min( mindat, dmin )
    maxdat = max( maxdat, dmax )
  return mindat, maxdat

#--------------------------------------------------------------------------------------------------

def plot_fitspace( outname, datname, **kwargs ):
  # Open the input file and get some general stats useful to all plots.
  # Call the routine for the desired type of plot.

  settings = kin_load_settings( datname, 0 )
  settings.update( kwargs )

  ptype = settings.get( "plot_type", "1d" )

  if ptype == '1d':
    _plot_fitspace_1d( outname, datname, **settings)

  elif ptype == 'normrates':
    _plot_fitspace_normrates( outname, datname, **settings )

  else:
    # 2d/3d grids
    grids = scan_grids( datname )
    if len( grids ) > 0:
      palette_name = datname.replace( ".txt", ".palette.txt" )
      print( "datname,palette_name = " + str([datname,palette_name]) ) 
      if 'palette' in palette_name and os.path.isfile( palette_name ):
        colormap = get_fs_colormap( palette_name )
        settings['colormap'] = colormap
        print( "using custom palette in " + palette_name )
      if ptype == 'contourf1024':
        settings['contour_n'] = 1024
        settings['contour_bounds'] = False

      _plot_fitspace_2d( outname, grids, **settings )
    else:
      print( "No grids to plot." )



#--------------------------------------------------------------------------------------------------

if __name__ == "__main__":

  argc = len( sys.argv )
  if argc <= 2:
    print( "Usage: python plotFitspace.py <eps_filename> <data_filename> <plot_type> <layout>" )
    print( "Ex:    python plotFitspace.py tryp.eps tryp.fsp.txt\n" )
  else:
    print( "\n===\n=== plotFitspace.py\n===\n" );

    ptype = "1d"
    layout = "default"
    if argc > 1:
      oname = sys.argv[1]
    if argc > 2:
      dname = sys.argv[2]
    if argc > 3:
      ptype = sys.argv[3]
    if argc > 4:
      layout = sys.argv[4]

    print( "dat is: " + dname )
    print( "out is: " + oname )
    print( "ptype is: " + ptype )
    print( "layout is: " + layout )

    if not dname or not os.path.isfile( dname ):
      print( "The data file '" + dname + "' is not found.\n" )
      quit()

    plot_fitspace( oname, dname, plot_type=ptype, layout=layout )

    print( "done.\n" )

#==================================================================================================
# Debug/Test
#==================================================================================================


def test_fs( filename, ptype ):
  plot_fitspace( "test.eps", filename, plot_type=ptype, layout="default", test=True)

#--------------------------------------------------------------------------------------------------

def get_grid_index( grids, name ):
  for i,g in enumerate(grids):
    if name == g['param1'] + ' vs ' + g['param2']:
      return i
  return None

def test_plot_grid( filename, palette_name=None, ptype="contourf", which="k-1 vs k+1" ):
  grids = scan_grids( filename )
  gridi = get_grid_index(grids,which)
  if gridi == None:
    print( "The grid '" + str(which) +"' was not found, using 1st grid." )
    gridi = 0

  fig,plots,cbax = _plots_layout_triangular( 2, '20x20', plot_type=ptype )
  ax = plots[0]['ax']

  p1,p2,dat,xlabel,ylabel = get_grid_info( grids, gridi )

  # colobar axes, colormap and range 
  kwargs = kin_load_settings( filename, 0 )
  kwargs['colorbar_axes'] = cbax
  kwargs['colormap'] = get_fs_colormap( palette_name ) if palette_name else plt.get_cmap( 'jet' )
  kwargs['vmin'],kwargs['vmax'] = get_grids_minmax( grids )
  kwargs['vmin'] = 0
    # we want contours and colors to stretch down to 0, even if the data does not.
  cb,cb_cfg = _config_colorbar( cbax, **kwargs )
  kwargs.update( cb_cfg )

  c = _plot_ptype( ptype, ax, p1, p2, dat, xlabel, ylabel, **kwargs )

  return { 'fig':fig, 'p1':p1, 'p2':p2, 'dat':dat, 'ylabel':ylabel, 'xlabel':xlabel }

