import sys
import os
import numpy as np
from plotconfig import *

#
# The Free Energy dialog exports the vertex locations for free-energy
#

#--------------------------------------------------------------------------------------------------

def plot_free_energy( outname, datname, axis1=None, axis2=None, **kwargs ):
  # The free-energy output can multiple multiple consecutive tables of data
  # that should be plotted on the same axes.  So we start by initializing
  # a figure, and then plotting each path as we encounter in while
  # scanning the file.

  #
  # Load cfg file if available.
  #
  cfg = kwargs.get( 'cfgname', None )
  if cfg and cfg.lower() != 'none':
    kin_load_cfg_file( cfg, kwargs, 1 )


  fig = kin_figure_init( figsize='9x7', init_single_plot=False ) 

  sub_pos = [ .14, .10, .82, .83 ]
  ax = fig.add_subplot( 1,1,1, position = sub_pos )

  minx = miny = sys.float_info.max
  maxx = maxy = sys.float_info.min

  color_index = 0
  with open(datname) as f:
    for index, line in enumerate(f):
      if 'ReactionCoordinate' in line:
        # here is a block of data that will end with a blank line
        header,dat = kin_load_data( datname, index+1, True, True )
        _minx,_maxx,_miny,_maxy = kin_plot( dat, [], ax, marker='-', color_index=color_index )  
        minx = min( minx, _minx )
        maxx = max( maxx, _maxx )
        miny = min( miny, _miny )
        maxy = max( maxy, _maxy )

      if 'Label' in line:
        # here is a block of labels with coordinates
        header,dat = kin_load_data( datname, index+1, False, True, False )
        # TODO: if this is python3, we need to convert the labels from bytes
        # to string, otherwise the EPS will have those silly b prefixes
        # before each label.  For now this is being done per label in the 
        # kin_plot_labels.  I tried doing a decode on those entries here,
        # but numpy arrays are special and have their own typing information
        # which they appear to enforce, i.e. overriding my attempt to set it
        # to a regular string.  I think you'd want to do a np.ndarray.astype
        # or something like that, which means you'd want this dat in a transposed
        # form of what it is.  Already we're currently asking for rows in
        # the call the kin_load_data, but only because the genfromtxt fn behaves
        # differently when we import lines from an array instead of lines from
        # a file -- see TODO comments in kin_load_data. We'll come back to this.
        _miny = kin_plot_labels( dat, ax, color_index=color_index, fontsize=kwargs.get('annotate_fontsize',9) )  
        miny = min( miny, _miny )
        color_index += 1


  #
  # Create Format axes, hide xlabels, hide negative ylabels
  # 
  kin_format_axes( ax, minx, maxx, miny, maxy, axis1, axis2, -99, False, True )
  plt.setp( ax.get_xticklabels(), visible=False )
  ax.tick_params( axis='x', direction='in' )
    # don't show arbitrary reaction coordinate labels, and tick marks point in.

  ylabels = ax.get_yticklabels()
  ym = ax.yaxis.get_major_ticks()
  yticks = ax.get_yticks()

  for i,label in enumerate(ylabels):
    if yticks[i] < 0:
      plt.setp( ylabels[i], visible=False )
      plt.setp( ym[i], visible=False )
        # making ym hidden does hide the label also, but it doesn't remove
        # the label from some kind of layout computation, so the axis label
        # text is spaces further away than it should be if I don't set both
        # of these visible=False.

  #
  # Save EPS & PNG
  #
  fig.savefig( outname, format='eps' )
  fig.savefig( outname.replace(".eps",".png"), format='png', dpi=300 )


#--------------------------------------------------------------------------------------------------

if __name__ == "__main__":

  argc = len( sys.argv )
  if argc <= 2:
    print( "Usage: python plotFreeEnergy.py <eps_filename> <data_filename> <axis1Label> <axis2Label> <cfgfilename>" )
    print( "Ex:    python plotFreeEnergy.py tryp.eps tryp.txt 'Reaction Coordinate' 'kcal/mol'\n" )
  else:
    print( "\n===\n=== plotFreeEnergy.py\n===\n" );

    axis1 = None
    axis2 = None
    cfgname = None
    if argc > 1:
      oname = sys.argv[1]
    if argc > 2:
      dname = sys.argv[2]
    if argc > 3:
      axis1 = sys.argv[3]
    if argc > 4:
      axis2 = sys.argv[4]
    if argc > 5:
      cfgname = sys.argv[5]

    print( "dat is: " + dname )
    print( "out is: " + oname )
    print( "axis1 is : " + str( axis1 ) )
    print( "axis2 is : " + str( axis2 ) )
    print( "cfgname is : " + str( cfgname ) )

    if not dname or not os.path.isfile( dname ):
      print( "The data file '" + dname + "' is not found.\n" )
      quit()

    plot_free_energy( oname, dname, axis1, axis2, cfgname=cfgname )

    print( "done.\n" )

