import sys
import os
import numpy as np
from plotconfig import *

#
# When analytic fits to a series are computed, it is possible to plot rates and amplitudes vs the 
# series variable, which is often concentration, but may also be temperature, voltage, or
# solvent concentration.  Throughout this code we always refer to "rate vs concentration" but
# it is understood that "concentration" may really be one of these other quantities.
#

#--------------------------------------------------------------------------------------------------

def plot_rate_v_conc( outname, datname, paramType="b", axis1=None, axis2=None, plot_sigma=False ):
  # scan the input file looking for the table that gives conentrations and rates.
  tableLine=-1
  with open(datname) as f:
    for index, line in enumerate(f):
      if 'Series Summary' in line:
        tableLine = index + 2 
          # this is the 0-based index where the table begins (column labels)
        break

  if tableLine == -1:
    print( "A 'Series Summary' entry was not found in the file " + datname + "\n" )
    return

  header,dat = kin_load_data( datname, tableLine+1 )
  labels = header[-1]

  #
  # special case for rate vs amplitide: we prefer to plot the absolute value
  # of the amplitude terms.
  #
  if paramType=='a':
    for i,la in enumerate(labels):
      if label_match( la, 'a\\d' ):
        dat[i] = np.absolute( dat[i] )

  #
  # Create figure/axes, plot, format axes, & save EPS, PNG
  # 
  fig,ax = kin_figure_init( figsize='9x7' )
  match_col = 'a\\d' if paramType=='a' else paramType
    # because amplitude-type params are always aN,
    # but in the case of a linear fit, the rate param can be just 'b' and not bN.
    # NOTE: if you want to plot the linear 'c' terms in the amp-vs-conc plot,
    # use 'a\\d|c' as the regexp.

  minx,maxx,miny,maxy = kin_plot( dat, labels, ax, label_match=match_col, plot_sigma=plot_sigma ) 
  minx = 0
  maxx = maxx * 1.1
  kin_format_axes( ax, minx, maxx, miny, maxy, axis1, axis2 ) 
  fig.savefig( outname, format='eps' )
  fig.savefig( outname.replace(".eps",".png"), format='png', dpi=300 )



#--------------------------------------------------------------------------------------------------

if __name__ == "__main__":

  argc = len( sys.argv )
  if argc <= 2:
    print( "Usage: python plotRateVsConc.py <eps_filename> <data_filename> <axis1Label> <axis2Label> <paramType> <plotError>" )
    print( "Ex:    python plotRateVsConc.py tryp.eps tryp.txt 'Concentration (uM)' 'Rate (1/s)' b 1\n" )
  else:
    print( "\n===\n=== plotRateVsConc.py\n===\n" );

    axis1 = None
    axis2 = None
    paramType = "b"
      # paramType refers to the name of the params - rates are always "b" and amplitudes are always "a",
      # which is how they are presented in analytic fits of series in the KinTek Explorer software.
    plotError = 0
    if argc > 1:
      oname = sys.argv[1]
    if argc > 2:
      dname = sys.argv[2]
    if argc > 3:
      axis1 = sys.argv[3]
    if argc > 4:
      axis2 = sys.argv[4]
    if argc > 5:
      paramType = sys.argv[5]
    if argc > 6:
      plotError = int(sys.argv[6])

    print( "dat is: " + dname )
    print( "out is: " + oname )
    print( "axis1 is : " + str( axis1 ) )
    print( "axis2 is : " + str( axis2 ) )
    print( "pType is : " + paramType )
    print( "plErr is : " + str( plotError ) )

    if not dname or not os.path.isfile( dname ):
      print( "The data file '" + dname + "' is not found.\n" )
      quit()

    plot_rate_v_conc( oname, dname, paramType, axis1, axis2, plotError )

    print( "done.\n" )

