import matplotlib.pyplot as plt
from matplotlib import use
use('agg') # use agg backend so we don't need to worry about GUI
import numpy as np
import sys
import re
import json
import math
import struct

#
# Colors taken from the KinTek C++ software.  Note yellow colors (#4) are 
# commented out per Ken's request.  (Update: they are made darker).
#

plotColors = [
  '#FF0000',
  '#00FF00',
  '#0000FF',
  '#FFC800',   # Note: this is darker than used by the c++ program to look good on WHITE eps plots.
  '#00FFFF',
  '#FF00FF',
  '#00AA00',
  '#9966FF',
  '#FF6600',
  '#3399FF',
  '#FF99FF',
  '#009966'
]

#---------------------------------------------------------------------

def pprintd( dict ):
  # pretty-print a dictionary, for debugging
  return json.dumps( dict, sort_keys=True, indent=2 )

#---------------------------------------------------------------------

def wh_from_figsize( figsize ):
  p = re.compile( "^(\d+)x(\d+)$" )
  m = p.match( figsize )
  if m:
    pw = int( m.group(1) )
    ph = int( m.group(2) )
    return pw,ph
  return 0,0

def figsize_from_wh( w, h ):
  return str(w) + 'x' + str(h)

#---------------------------------------------------------------------

def kin_figure_init( figsize, init_single_plot=True, **kwargs ):
  #
  # Set KinTek defaults for single-column publication figure
  #
  plt.rcParams['ps.fonttype'] = 42
  plt.rcParams['ps.useafm'] = True
  plt.rcParams['pdf.use14corefonts'] = True
    # after lots of experimentation, this seems to do what we want, which is have text exported
    # not as generic paths but as text with font information so that it can easily be edited
    # in Illustrator.  I'm not sure all of these options are necessary.  On OSX, this results
    # in Helvetica for the default sans-serif, which is what we like.

  plt.rcParams['mathtext.default'] = 'regular'
    # we use mathtext strings to do super/subscript, and would like the font to
    # remain the same sans-serif we use for everything else.

  plt.rcParams['font.family'] = 'Times' if kwargs.get('fontserif',None) else 'sans-serif'
  plt.rcParams['axes.linewidth'] = 0.5
  plt.rcParams['lines.linewidth'] = 0.666
    # Just a special marker value so I can see where this shows up in plots,
    # typically any plotting specifies it's desired linewidth.

  sub_pos = [0.0,0.0,1.0,1.0]
    # these values determine how the subplot is poitioned in the 1x1 normalized
    # space of the figure.  Below this is adjusted based on the dimension of the 
    # plot and the font size used for axis labels such that the labels are not
    # clipped but whitespace around the figure is minimized.

  plt.rcParams['font.size'] = kwargs.get( 'font_size', 11 )
  plt.rcParams['xtick.labelsize'] = kwargs.get( 'ticklabel_size', plt.rcParams['font.size']-2 )
  plt.rcParams['ytick.labelsize'] = kwargs.get( 'ticklabel_size', plt.rcParams['font.size']-2 )

  if figsize == '9x7':
    plt.rcParams['figure.figsize'] = ( 9/2.54, 7/2.54 )
       # 9x7 cm, used for experiment exports & 2D data export
    sub_pos = [ .14, .15, .82, .85 ]
  elif figsize == '10x10' or figsize == '12x10':
    w,h = wh_from_figsize( figsize )
    plt.rcParams['figure.figsize'] = ( w/2.54, h/2.54 )
      # 10x10cm, used for 2D projections of 3d spectra (wavelen or time mode)
    plt.rcParams['font.size'] = 12
    plt.rcParams['xtick.labelsize'] = 10
    plt.rcParams['ytick.labelsize'] = 10
    if w == 10:
      sub_pos = [0.11, 0.12, 0.86, 0.88]
    else:
      sub_pos = [0.11, 0.12, 0.76, 0.88]
        # the 12x10 is a special format to allow space for a legend outside
        # of the axes.  The axes should still have a roughly square aspect.
  elif figsize == '20x20':
    plt.rcParams['figure.figsize'] = ( 20/2.54, 20/2.54 )
      # 20x20 cm, for 3d spectra plots in "ortho" mode
    plt.rcParams['font.size'] = 14
    plt.rcParams['xtick.labelsize'] = 12
    plt.rcParams['ytick.labelsize'] = 12
    sub_pos = [0.05, 0.045, 0.93, 0.955]
  else:
    # The above are specific types of plots that we want tight control
    # over, but in some cases a custom size may be asked for (e.g. fitspace
    # plots where you may have lots of plots in one large figure)
    w,h = wh_from_figsize( figsize )
    plt.rcParams['figure.figsize'] = ( w/2.54, h/2.54 )
      # TODO: fontsize etc. Should caller compute this?
      # For now the default smallest size will get used, which is
      # appropriate for current usage in which custom figsize is
      # only used when putting lots of smaller subplots on a figure.
  
  plt.rcParams['legend.fontsize'] = plt.rcParams['font.size']
  plt.rcParams['axes.labelsize'] = plt.rcParams['font.size']
  plt.rcParams['axes.titlesize'] = 1.5*plt.rcParams['font.size']

  plt.clf()   
    # perhaps unnecessary since we are creating a new fig anyway.

  fig = plt.figure()
  if init_single_plot:
    axe = fig.add_subplot( 1,1,1, position = sub_pos )
    return fig,axe

  return fig

#---------------------------------------------------------------------

def kin_thin_ticklabels( ax, maxlabels, thinx=True, thiny=True, keep_blank_ticks=True ):

  process = [ thinx, thiny ]

  fns = [ { 'get_ticks':ax.get_xticks, 'set_ticks':ax.set_xticks, 'set_ticklabels':ax.set_xticklabels },
      { 'get_ticks':ax.get_yticks, 'set_ticks':ax.set_yticks, 'set_ticklabels':ax.set_yticklabels } ]

  for i,fn in enumerate(fns):
    if process[i]:
      ticks = fn['get_ticks']()
      if len(ticks) > maxlabels:
        if i==0 and ticks[0] < 0:
          ticks = ticks[1:]
            # discard first negative x-label, usually not printed anyway 
        tickvals = []
          # we'll convert them to ints where possible to avoid unnecessary decimal output
        ticklabels = []
        for it,t in enumerate(ticks):
          tickval = int(t) if int(t)==t else t

          if it % 2:
            # skip these, but keep ticks without labels if requested
            if keep_blank_ticks:
              tickvals.append( tickval )
              ticklabels.append( '' )
            continue

          # keep every other tick and label
          tickvals.append( tickval )
          ticklabels.append( tickval )
        
        fn['set_ticks']( tickvals )
        fn['set_ticklabels']( ticklabels )



#---------------------------------------------------------------------

def kin_normalize_axis( values, label ):
  #
  # divide the minv and maxv by a value which is then included as a term in the label.
  # e.g. if minv=1000 and maxv=1e6, divide out 1000, and include the /1000 in the label.
  # This was written for chi2 fitspace plots where sometimes a rate has a range e.g.
  # [ 1000, 8000] and you'd prefer the xticks to be range [1,8 ] and the label to say
  # k+1/1e3
  #
  factor = 1
  # m = np.median( values )
    # the median works well for most cases.  But in some situations like 8oxodg, k+3 is 
    # sampled heavily at very small values but then stretches to quite a larger value, 
    # so we use a 10^-2 exponent where it's not really necessary.
  m = np.average( values )
  e = math.floor( math.log10( m ) )
  if abs(e) >= 2.0:
    factor = math.pow( 10, e )
    values = np.divide( values, factor )
    label = label.replace( '$', '' )
    label = '$' + label + '/(10^{' + str(int(e)) + '})$'

# print( "   average " + str(m) + " resulted in e=" + str(e) + " and label " + str(label) )
  
  return values, label, factor

#---------------------------------------------------------------------

def adjust_y_axis_bounds( miny, maxy, prefer_miny0=True, expand_y=True ):
  delta_y = ( maxy - miny ) * .1
  if expand_y:
    maxy += delta_y

  if prefer_miny0 and miny >= 0 and miny-delta_y < 0:
    # if miny is close enough to 0 that subtracting delta_y would
    # place it below 0, then just set it to 0 so that the x-axis will
    # line up with y=0.
    miny = 0
  elif expand_y:
    miny = miny - delta_y

  return miny, maxy

def kin_format_axes( ax, minx, maxx, miny, maxy, axis1=None, axis2=None, logx=-99, prefer_miny0=True, expand_x=False, expand_y=True ):
  #
  # Set axes bounds with some optional padding as preferred by Ken. 
  # Also sets labels.  logx magic value -99 means NOT log axis.
  #
  miny, maxy = adjust_y_axis_bounds( miny, maxy, prefer_miny0, expand_y )

  if expand_x:
    delta_x = ( maxx - minx ) * .05
    #minx -= delta_x
    maxx += delta_x

  if logx != -99:
    ax.set_xscale( 'log' )
    minx = pow( 10, logx )
    #print( "kin_format_axes: logx is " + str(logx) + " and minx computed as " + str(minx) );

  ax.axis( [ minx, maxx, miny, maxy ] )

  ax.ticklabel_format( style='sci', scilimits=(-3,4), axis='both' if logx==-99 else 'y' )
    # use scientific notation for numbers outside the range [ .001, 10000 ]
  
  #
  # set axis labels
  #
  if axis1:
    ax.set_xlabel( translate_symbols(axis1) )
      # labelpad is to move the label up some
  if axis2:
    ax.set_ylabel( translate_symbols(axis2), labelpad=2 )
      # labelpad is to move the label to the left some


#---------------------------------------------------------------------

def kin_load_cfg_file( filename, to_hash, verbose=0 ):
  cfgoptions = kin_typecast_settings( kin_load_settings( filename, 0 ) )
  to_hash.update( cfgoptions )
  if verbose==1:
    print( "Loaded config from file '" + filename + "':" )
    for k,v in iter( cfgoptions.items() ):
      print( "  " + k + " " + str(type(v)) + " --> " + str(v) )
  

def kin_load_settings( filename, header_rows ):
  # loads key-value pairs from file to dict
  settings = {}
  p = re.compile( "^(\S+)\s+(.+)?$" )
  with open( filename, 'r' ) as f:
    for i,line in enumerate( f ):
      if i<header_rows: 
        continue

      m = p.match( line )
      if( m ):
        settings[ m.group(1) ] = m.group(2)
      else:
        break

  return settings

def kin_typecast_settings( _settings ):
  # cast the settings we know that we load into the correct type
  # because if for example line_width is read as a string above,
  # it will cause problems downstream in matplotlib.

  casted = {}
  for k,v in iter( _settings.items() ):
    if k == 'linewidth' or k == 'marker_size' or k == 'font_size' or k == 'line_brightness' or k == 'time_max' or k=='x0' or k=='x1' or k=='y0' or k=='y1':
      casted[k] = float( v )
    elif k == 'fontserif' or k == 'data_lines' or k == 'plot_residuals' or k == 'norm_residuals' or k == 'plot_sigma' or k == 'show_fits':
      casted[k] = int( v )
    elif k == 'color_offset' or k == 'data_indices' or k == 'sim_indices' or k == 'res_indices':
      casted[k] = tuple( int(item) for item in v.split(',') if item.strip() )
    elif k == 'custom_colors':
      casted[k] = tuple( item for item in v.split(',') if item.strip() )
    elif k == 'data_sigmas':
      casted[k] = tuple( float(item) for item in v.split(',') if item.strip() )

  _settings.update( casted )
  return _settings

#---------------------------------------------------------------------

def get_sigma_from_header( header ):
  # header is header as returned from kin_load_data, so something like
  # [['Sigma', '1.5'],['Time', 'S1']]
  # 
  # Looks through lines in header for definition of sigma as exported
  # by KinTek Explorer.  This is either a single value to be applied
  # to all data, or a number of values, 1-per column of data.

  sigma_avg = None
  sigma_cols = None
  for line in header:
    if line[0].startswith( "Sigma" ):
      sigma_cols = []
      hasNumeric = False
      cols = line[1:]
      for col in cols:
        try:
          val = float( col )
          hasNumeric = True
        except ValueError:
          val = 0.0
        sigma_cols.append( val )
      if len(sigma_cols) == 1:
        sigma_avg = sigma_cols[0]
        sigma_cols = None
      if not hasNumeric:
        sigma_avg = None
        sigma_cols = None
      else:
        return sigma_avg, sigma_cols
          # return first Sigma line that has values

  return sigma_avg, sigma_cols


#---------------------------------------------------------------------

def kin_load_data( filename, header_rows, import_cols=True, break_on_blank_line=False, force_float=True, **kwargs ):
  # Returns array of header rows and a separate matrix with rows or columns
  # of data, e.g. data[0] is the first col of the file if import_cols is True.
  #
  # If break_on_blank_line=False, the filename is fed to numpy.loadtxt
  # and it is told to skip the header_rows.  If break_on... is True,
  # then we scan the lines ourselves into an array and stop when we
  # hit a blank line, and then feed the lines to numpy.loadtxt

  header = []
    # each entry will be an array corresponding to column titles
  datlines = []
  datsource = filename

  delimiter = kwargs.get( 'delimiter', None )

  fp = open( filename,'r' )
  if fp:
    for i, line in enumerate(fp):
      if i < header_rows:
        header.append( line.strip().split( delimiter ) )
      elif break_on_blank_line:
        if not line.isspace():
          add = line.strip()
          if sys.version_info > (3,0):
            add = bytes(add,"utf-8")
            # bytes is required for parsing by numpy genfromtxt.  This is
            # not an issue in python2
          datlines.append( add )
        else:
          break
      else:
        break

    fp.close()

    if break_on_blank_line:
      datsource = datlines
      header_rows = 0
      # TODO: when you pass genfromtxt an array of lines instead of a filename,
      # it formats the resulting matrix differently than you'd expect - it
      # effectively ignores the unpack parameter and returns you matrix rows
      # corresponding to the lines.  To fix this we'd need to transpose the
      # input here if import_cols is true.
      #
      # NOTE: I don't think this is true.  There must have been some other bug
      # causing this when I made this note, so I'm leaving it for now. But 
      # I'm using break_on_blank_line in the scan_grids() for plotFitspace,
      # and the import_cols works as you expect it to.

    gen_args = {}
    gen_args['comments'] = kwargs.get( 'comments', '^' )
      # TODO: I just put ^ as comments because we need a char that will never show up in species names
    gen_args['dtype'] = float if force_float else None
    gen_args['unpack'] = import_cols
    gen_args['delimiter'] = delimiter
    gen_args['missing_values'] = kwargs.get( 'missing_values', None )
    gen_args['filling_values'] = kwargs.get( 'filling_values', None )
      # TODO: what about NaN?
      
    data = np.genfromtxt( datsource, skip_header=header_rows, **gen_args );
      # use genfromtxt to allow loading of string data in columns as well.

  # TODO: should we convert byte sequences to strings so that if these are printed
  # they don't have the silly b
  return ( header, data )

#---------------------------------------------------------------------

def kin_load_data_3d( filename, import_cols=True, skip_lines=1, break_on_blank_line=False, force_float=True, **kwargs ):
  # load data as exported by our program
  # returns (time,wavelengths,data)
  # where time,wavelengths are 1d vector, data is 2d matrix

  ( header, data ) = kin_load_data( filename, skip_lines, True, break_on_blank_line, **kwargs )
  time = [ float(x) for x in header[skip_lines-1][1:] ]
  wavelengths = data[0]
  data = np.delete( data, 0, 0 )
  if not import_cols:
    data = np.transpose( data )
  return ( time, wavelengths, data )


#---------------------------------------------------------------------
# TODO: if there is a speed issue here (e.g. if get_plot_color is 
# called many times in the course of some plotting operation) then 
# consider converting our color arrays at top to rgb triplets instead.

def _to_rgb( hextriplet ):
  return tuple(int(hextriplet[i:i+2], 16) for i in (0, 2 ,4))

def _to_hex_triplet( rgb ):
  return '%02x%02x%02x' % rgb

def get_plot_color( index, brightness=1, color_offset=None, custom_colors=None ):
  # The plot colors as used in the KinTek program are given in plotColors
  # at the top of this file.  
    #
  # Here 'index' refers to the nth trace being plotted in the figure, and we
  # typically index directly into this list of colors.
  #
  # It is sometimes desireable to specifiy a color offset to apply at a location,
  # which would affect the "effective index" for any index value >= to the 
  # offset location.  color_offset may be passed as a kwa
  #
  # brightness of the colors may be adjusted [0,1]

  if color_offset:
    # color offset is a tuple giving offset location & offset size
    if index >= color_offset[0]:
      index += color_offset[1] - color_offset[0]

  numPlotColors = len( plotColors )
  i = index % numPlotColors
  color = plotColors[i]

  if custom_colors and len(custom_colors)>i and custom_colors[i] != "0":
    print( "Setting color to custom_color " + custom_colors[i] )
    color = custom_colors[i]

  if brightness != 1:
    rgb = _to_rgb( color[1:] )
    color = '#' + _to_hex_triplet(  tuple(int(brightness * v) for v in rgb) )
  return color

#---------------------------------------------------------------------

def get_marker_type_and_size( data_length ):
  marker_size = 2.0
  marker_type = '.'
  if data_length < 20:
    marker_size = 4
    marker_type = 'o'
  elif data_length < 50:
    marker_size = 3.0
    marker_type = 'o'
  elif data_length < 99:
    marker_size = 2.5
    marker_type = 'o'
  
  # marker_size values are from matlab and matplotlib is different.
  # matplotlib marker_size speaks to the AREA of the marker, so to double the 'width'
  # of a marker you need to mult the marker_size by 4

  return ( marker_type, marker_size )

#---------------------------------------------------------------------

def get_linewidth_for_line_with_data( sigma, delta_y, fig ):
  # fig is a param because I want to be able to query for the size of
  # the y-axis in mm or points, but for now I'm going to user the dim
  # that we are using.
  y_axis_length_in_points = 53.8 / .353
  linewidth = sigma / delta_y * y_axis_length_in_points
  if linewidth < 0.25:
    linewidth = 0.25
  elif linewidth > 2.0:
    linewidth = 2.0
  return linewidth


#---------------------------------------------------------------------

def label_match( label, match_expr ):
  p = re.compile( match_expr, re.IGNORECASE )
  return p.match( label )

def label_index( label ):
  p = re.compile( "^S(\d+)$" )
  m = p.match( label )
  if m:
    return int( m.group(1) )
  return -1

def unique_labels( labels ):
  return len(labels) == len(set(labels))

#---------------------------------------------------------------------

def parameter_label( param ):
  # convert the plain string to a $-delimited mathtext string if
  # the param has characters that we're going to subscript.
  p = re.compile( "^\s*k([\+-]\d+)\s*$" )
  m = p.match( param )
  if m:
    param = '$k_{' + m.group(1) + '}$'
  return param

#---------------------------------------------------------------------

def kin_check_times_ascending( dat, labels ):
  dat_cols = dat.shape[0]
  dat_len = len(dat[0])

  for col in range( 0, dat_cols ):
    if labels[col].lower() == "time":
      # check that this column is monotonic increasing
      time_col = dat[col]
      for row in range( 0, dat_len-1 ):
        if not np.isfinite( time_col[row] ):
          print( "Warning: not finite time in row,col,val: " + str([row,col,time_col[row]]) )
        if not np.isfinite( time_col[row+1] ):
          print( "Warning: not finite time in row,col,val: " + str([row+1,col,time_col[row+1]]) )
        if time_col[row+1] < time_col[row]:
          print( "Warning: not ascending time in row,col,va1,va2: " + str([row+1,col,time_col[row],time_col[row+1]]) )


#---------------------------------------------------------------------

def kin_concat_data( dat, labels ):
  # concatenate all data (non-"time" columns) into a single array.
  arys = []
  dat_cols = dat.shape[0]
  for col in range( 0, dat_cols ):
    if labels[col].lower() != "time":
      arys.append( dat[col] )

  return np.concatenate( arys )


#---------------------------------------------------------------------
# This was written to scan data that will be plotted using the logic
# of kin_plot, and return the min/max y values that will be plotted.
# This was needed so that error bars can be manually clipped to those
# min/max values because if you don't, and error bars go off the chart,
# even if they are not visible because the axes object clips them, it
# causes problems in Adobe Illustrator when you try to move the figure.

def kin_minmaxy_data( dat, labels, argv ):
  dat_cols = dat.shape[0]

  miny = sys.float_info.max
  maxy = sys.float_info.min

  for col in range( 1, dat_cols ):
    label = labels[col].lower()
    if label != "sigma" and label != "time":
      if label_match( labels[col], argv['label_match'] ):
        miny = min( miny, np.nanmin(dat[col]) )
        maxy = max( maxy, np.nanmax(dat[col]) )

  return miny, maxy

#---------------------------------------------------------------------

def kin_plot( dat, labels, ax, **kwargs ):
  # An attempt at a general purpose plotter that 
  # 1. allows you to set marker & marker_size/line_width
  # 2. allows to to specify which columns should be plotted by matching the labels
  # 3. knows to look for "time" in labels to use a new set of x-coordinates
  # 4. knows to look for "sigma" in labels and optionally plot this as error bars
  # 5. returns min/max values for each axis to the caller to allow setting bounds

  argv = {
    'marker_size'   : 2,
    'linewidth'   : 1.5,
    'label_match' : '',
    'plot_sigma'  : False,
    'color_offset'  : None,
    'color_indices' : None,
    'custom_colors' : None,
    'color_index'   : None
  }
  argv.update( kwargs )

  plot_colors = [] 
  error_colors = []
  if 'colors' in argv:
    # if multiple colors specified, we'll pick one from the list for each trace
    plot_colors = argv['colors'].split()
  if 'colors_error' in argv:
    error_colors = argv['colors_error'].split()
  
  plot_color = argv.get( 'color', 'b' )
    # if one color is specified instead, that color will be used for each trace
    # (if no colors specified, defaults from KinTekExplorer will be used)
  line_brightness = argv.get( 'line_brightness', 1 )

  # to start we're assuming column-data where labels is dat.cols int length

  dat_cols = dat.shape[0]
  dat_len = len(dat[0])

  if not 'marker' in kwargs:
    argv['marker'],marker_size = get_marker_type_and_size( dat_len )    
    if not 'marker_size' in kwargs:
      # for some plots the c++ software passes in the marker_size, but if not,
      # then use the value based on dat_len.
      argv['marker_size'] = marker_size

  if not labels:
    labels = []
  labels = labels + [''] * ( dat_cols - len(labels) )

  # Returns max x and y values to set axis ranges
  #
  minx = miny = sys.float_info.max
  maxx = maxy = sys.float_info.min

  if argv['plot_sigma']:
    # If we will plot error bars, we are going to clip them such that
    # those that would extend WAY outside the figure only extend a little
    # bit outside the figure.  This is because Adobe Illustrator has a hard
    # time with EPS files with error bars that stick way out past page
    # boundaries.
    pre_miny, pre_maxy = kin_minmaxy_data( dat, labels, argv )
    pre_miny, pre_maxy = adjust_y_axis_bounds( pre_miny, pre_maxy )
      # adjust like we normally do to expand the y axis a bit.
    deltay = ( pre_maxy - pre_miny ) * .1
    pre_miny -= deltay
    pre_maxy += deltay
      # and adjust a bit more so that error bar actually goes outside of
      # figure, so that we don't see the end-caps and its obvious this
      # error goes out of bounds.


  # Plot - assumption here is that the first col must always be "time" equivalent,
  # that is, indicate the domain.
  #
  domain_labels = [ "time", "conc" ]
  cols_plotted = 0
  time_col = 0
  color_by_label = unique_labels( labels )
  for col in range( 1, dat_cols ):
    if labels[col].lower() != "sigma":

      col_label = labels[col].lower()
      if col_label in domain_labels:
        time_col = col
        continue

      if label_match( labels[col], argv['label_match'] ):
        marker = argv['marker']
        if 'colors' in argv:
          plot_color = plot_colors[ cols_plotted % len(plot_colors) ]
        elif 'color' not in argv:
          if argv['color_index']:
            color_index = argv['color_index']
          else:
            observable_index = label_index( labels[col] ) - 1 if color_by_label else -1
              # get observable index from label if possible, unless there are duplicate labels (which is probably a series)
            color_by_col = observable_index if observable_index >= 0 else cols_plotted
              # if we successfully got an observable index from the label, use it, otherwise use column number
            color_index = argv['color_indices'][cols_plotted] if argv['color_indices'] else color_by_col
              # if we have specific color_indices per column, use it, else use the color index we've previously arrived at.
          
          plot_color = get_plot_color( color_index, line_brightness if marker=='-' else 1, argv['color_offset'], argv['custom_colors'] )

        plotted_sigma = False

        next_col_is_sigma = col<dat_cols-1 and labels[col+1].lower()=="sigma"
        avg_sigma = argv.get( 'sigmaAvg', None )
        col_sigma = argv.get( 'sigmaCols', None )
        plot_sigma = argv['plot_sigma'] and (next_col_is_sigma or avg_sigma or (col_sigma and len(col_sigma)==dat_cols-1) )

        if plot_sigma:
          if next_col_is_sigma:
            to_plot = np.array( [ dat[time_col], dat[col], dat[col+1], dat[col+1] ] ).transpose()
          else:
            sigma_val = avg_sigma if avg_sigma else col_sigma[col-1]
              # col-1 because the initial Time column has no sigma value
            sigmas = np.empty( len(dat[time_col]) )
            sigmas.fill( sigma_val )
            to_plot = np.array( [ dat[time_col], dat[col], sigmas, sigmas ] ).transpose()

          to_plot = to_plot[ ~np.isnan(to_plot).any(1) ]
          
          #
          # For exported plots, ensure that the error bars only exceed the figure
          # boundaries by a small amount, otherwise you may get errors when trying
          # to move the figure in Adobe Illustrator.
          #
          for i in range( 0, len(to_plot) ):
            row = to_plot[i]
            bot = row[1] - row[2]
            top = row[1] + row[3]
              # note that row[3] is same as row[2] unless we modify below...
            if top > pre_maxy:
              row[3] = pre_maxy - row[1]
            if bot < pre_miny:
              row[2] = row[1] - pre_miny

          to_plot = to_plot.transpose()

          ecolor = plot_color;
          if 'colors_error' in argv:
            ecolor = error_colors[ cols_plotted ]

          linestyle = 'solid'
          if marker=='-':
            # Newer matplotlib balks at a marker of '-' which used to just mean plot lines.
            marker = None
          else:
            # Prevent markers connected by lines when calling errorbar
            linestyle = 'None'

          ax.errorbar( to_plot[0], to_plot[1], yerr=[ to_plot[2], to_plot[3] ], marker=marker, ms=argv['marker_size'],
            linestyle=linestyle, linewidth=argv['linewidth'], color=plot_color, ecolor=ecolor, capsize=3, zorder=-100 )
          plotted_sigma = True
        else:
          to_plot = np.array( [ dat[time_col], dat[col] ] ).transpose()
          to_plot = to_plot[ ~np.isnan(to_plot).any(1) ].transpose()
          ax.plot( to_plot[0], to_plot[1], marker, ms=argv['marker_size'], linewidth=argv['linewidth'],
            color=plot_color, label=str(labels[col]), zorder=argv.get('zorder',0) )

        cols_plotted += 1
        maxerr = 0
        # if plotted_sigma:
        #   maxerr = np.amax( dat[col+1] )
        # Ken prefers we not do this because in some cases the errors on a point or two are
        # so large you lose the ability to see the trend in the actual data.
        minx = min( minx, np.nanmin(dat[time_col])  )
        maxx = max( maxx, np.nanmax(dat[time_col]) )
        miny = min( miny, np.nanmin(dat[col]) - maxerr )
        maxy = max( maxy, np.nanmax(dat[col]) + maxerr )

  return minx, maxx, miny, maxy

def kin_plot_fill_same_labels( dat, labels, ax, **kwargs ):
  # A very specific function to fill in the area between traces that have the 
  # same labels.  Some very specific assumptions:

  # 1. All columns must share the same x coordinates from column 0

  facecolor = kwargs.get( 'facecolor', '0.85' )
  edgecolor = kwargs.get( 'edgecolor', '0.75' )
  linewidth = kwargs.get( 'linewidth', '.2')
  dat_cols = dat.shape[0]

  label_processed = {}

  for col in range( 1, dat_cols ):
    if not label_processed.get( labels[col], None ):
      label_processed[ labels[col] ] = True
      for other in range( col+1, dat_cols ):
        if labels[other] == labels[col]:
          # same label, so do a fill between these traces
          ax.fill_between( dat[0], dat[col], dat[other], facecolor=facecolor, edgecolor=edgecolor, linewidth=linewidth, zorder=-100 )
            # zorder to ensure this plots below other lines/data

#---------------------------------------------------------------------

def kin_plot_labels( labels, ax, color_index, fontsize=9 ):
  # labels is a matrix in which col 0 contains the texts, and 
  # cols 1,2 contain the x,y coords at which to print them.
  miny = sys.float_info.max
  for row in labels:
    label = row[0]
    if sys.version_info > (3,0):
      # TODO: It would be cleaner to move this outside of this fn and 
      # process all of the labels after they are imported from the file.
      # The issue here is that in in python3 bytearrays and strings 
      # are strictly different so we get type errors in many places
      # after using np.genfromtxt.  Maybe we can tell genfromtxt to
      # treat a column as a string and not a bytearray.
      label = label.decode()

    textcolor = get_plot_color( color_index )
    ax.annotate( label, (row[1],row[2]), fontsize=fontsize, color=textcolor )
    miny = min( miny, row[2] )

  return miny


#---------------------------------------------------------------------

def translate_symbols( text ):
  # We initially tried to support translation to unicode, but this did not work for EPS exports,
  # though it did work if you just did a show() on the plot from python.
  #
  #t = text.replace( "\\mu", u"\u03bc" )
    # doesn't work for EPS export

  # The escaped mathtext approach seems to work for the exports.  We could allow special 
  # characters to be input in the kintek program, but right now we're just supporting the 
  # common case requested by Ken - the micro-molar legend text often used in sim plots:

  t = text.replace( "uM", "$\mu$M" )
  t = t.replace( "(us)", "($\mu$s)" )
    # and also for microseconds, but here require the parens, since the string "us" is perhaps
    # more likely to occur than the mixed case "uM"

  return t

#---------------------------------------------------------------------

def patch_EPS_with_greek_glyphs( filename, find=None, replace=None ):
  #
  # This is no longer used!  See translate_symbols above.  We did
  # this hacky patch_EPS when the unicode approach was failing,
  # but the escaped-mathtext approach seems to work for both 
  # export formats: http://matplotlib.org/users/mathtext.html
  #

  #
  # This is a workaround for matplotlib not correctly using greek 
  # postscript glyphs when exporting to EPS files.  It might work
  # if we used the "usetex" option, but we don't want to require
  # that the user have LaTex installed. 
  #
  #
  # Initially this was written to find specific combos and replace them.
  # But later I added the find/replace params, which only seek to 
  # replace specific text with other text.
  #
  # For the first method, it's like this:
  #
  # Look for occurences of specific combinations of characters so
  # that we can replace one of them with a greek glyph. 
  #
  # e.g. replace uM with (greekmu)M
  #
  # This relies on the standin character being the right size,
  # because we simply relace the glyph, as in:

  # 68.260000 2.070000 m /u glyphshow
  # 73.820000 2.070000 m /M glyphshow
  #
  # becomes:
  #
  # 68.260000 2.070000 m /mugreek glyphshow
  # 73.820000 2.070000 m /M glyphshow

  lines = []
  with open( filename ) as f:
    lines = f.readlines();
    for i,li in enumerate(lines):
      if not find:
        #
        # First method -- look for specific combos we always replace
        #
        if "/u glyphshow" in li and "/M glyphshow" in lines[i+1]:
          # we know a /u will never be on the last line of an EPS file so i+1 is reasonable
          lines[i] = li.replace( "/u", "/mugreek" )
        if "/r glyphshow" in li and "/slash glyphshow" in lines[i+1] and "/asterisk glyphshow" in lines[i+2]:
          # special case for r/* we emit when plotting residuals which have been normalized.
          # as above, we know we can safely index to i+1 and i+2 because of the format of an EPS file.
          lines[i+2] = lines[i+2].replace( "/asterisk", "/sigma" )
      else:
        #
        # second method -- look for find and replace it
        #
        if find in li:
          # special case for for residuals histogram file
          lines[i] = li.replace( find, replace )

  if len( lines ) > 0:
    with open( filename, 'w' ) as out:
      for li in lines:
        out.write( li )
